import {
  Component,
  OnInit,
  Output,
  EventEmitter,
  ViewChild,
  ElementRef,
  Input,
  NgZone
} from '@angular/core';
import { NgForm } from '@angular/forms';
import { FileUploadItem } from '@ucap-webmessenger/api';
import {
  FileUploadQueueComponent,
  SnackBarService,
  AlertSnackbarComponent,
  AlertSnackbarData
} from '@ucap-webmessenger/ui';
import { CommonApiService } from '@ucap-webmessenger/api-common';
import { NGXLogger } from 'ngx-logger';
import { FileUtil } from '@ucap-webmessenger/core';
import { TranslateService } from '@ngx-translate/core';

@Component({
  selector: 'ucap-chat-form',
  templateUrl: './form.component.html',
  styleUrls: ['./form.component.scss']
})
export class FormComponent implements OnInit {
  @Input()
  fileUploadQueue: FileUploadQueueComponent;

  @Output()
  send = new EventEmitter<string>();

  @Output()
  sendFiles = new EventEmitter<FileUploadItem[]>();

  @Output()
  toggleStickerSelector = new EventEmitter<void>();
  @Output()
  toggleTranslation = new EventEmitter<void>();

  @Output()
  clearView = new EventEmitter();

  @Output()
  clipboardPaste = new EventEmitter<ClipboardEvent>();

  @ViewChild('replyInput', { static: false })
  replyInput: ElementRef<HTMLTextAreaElement>;

  @ViewChild('fileInput', { static: false })
  fileInput: ElementRef<HTMLInputElement>;

  constructor(
    private translateService: TranslateService,
    private commonApiService: CommonApiService,
    private snackBarService: SnackBarService,
    private readonly ngZone: NgZone,
    private logger: NGXLogger
  ) {}

  ngOnInit() {}

  focus(): void {
    this.replyInput.nativeElement.focus();
  }

  onSend(event: Event) {
    event.preventDefault();

    this.send.emit(this.replyInput.nativeElement.value);
    this.replyInput.nativeElement.value = '';
  }

  onClickFileInput() {
    this.clearView.emit();
    this.fileInput.nativeElement.click();
  }

  onChangeFileInput() {
    const self = this;
    const fileList = this.fileInput.nativeElement.files;
    this.validUploadFile(fileList)
      .then(async result => {
        if (!result) {
          self.fileInput.nativeElement.value = '';
          return;
        } else {
          const fileUploadItems = FileUploadItem.fromFiles(fileList);

          if (!!self.fileUploadQueue) {
            self.fileUploadQueue.onFileSelected(fileUploadItems);
          }
          self.fileInput.nativeElement.value = '';
          self.sendFiles.emit(fileUploadItems);
        }
      })
      .catch(err => {
        self.fileInput.nativeElement.value = '';
        this.logger.error(err);
      });
  }

  onClickStickerSelector() {
    this.toggleStickerSelector.emit();
  }

  onClickTranslation() {
    this.toggleTranslation.emit();
  }

  onPasteReply(event: ClipboardEvent) {
    event.preventDefault();

    this.clipboardPaste.emit(event);
  }

  async validUploadFile(fileList: FileList): Promise<boolean> {
    let valid = true;

    const checkExt = this.commonApiService.acceptableExtensionForFileTalk(
      FileUtil.getExtensions(fileList)
    );
    if (!!checkExt) {
      this.ngZone.run(() => {
        this.snackBarService.openFromComponent<
          AlertSnackbarComponent,
          AlertSnackbarData
        >(AlertSnackbarComponent, {
          duration: 1000,
          verticalPosition: 'bottom',
          horizontalPosition: 'center',
          data: {
            html: this.translateService.instant(
              'common.file.errors.notSupporedType',
              {
                supporedType: checkExt.join(',')
              }
            )
          }
        });
      });
      valid = false;
      return valid;
    }

    const fakeMedia = await this.commonApiService.checkInvalidMediaMimeForFileTalkForFileList(
      fileList
    );
    if (!!fakeMedia) {
      this.ngZone.run(() => {
        this.snackBarService.openFromComponent<
          AlertSnackbarComponent,
          AlertSnackbarData
        >(AlertSnackbarComponent, {
          duration: 1000,
          verticalPosition: 'bottom',
          horizontalPosition: 'center',
          data: {
            html: this.translateService.instant(
              'common.file.errors.notAcceptableMime',
              {
                supporedType: fakeMedia.join(',')
              }
            )
          }
        });
      });
      valid = false;
      return valid;
    }

    return valid;
  }
}
